-- ============================================================= --
-- FOG CONTROL MOD
-- ============================================================= --
FogControl = {}
FogControl.name = g_currentModName
FogControl.path = g_currentModDirectory

FogControl.SETTINGS = {}
FogControl.CONTROLS = {}

addModEventListener(FogControl)

FogControl.menuItems = {
    'fogEnabled',
    'groundFogDensity',
    'heightFogDensity',
    'fogCoverage',
    'fogStartTime',
    'fogEndTime',
}

-- Default settings
FogControl.fogEnabled = true
FogControl.groundFogDensity = 0.005
FogControl.heightFogDensity = 0.6
FogControl.fogCoverage = 0.015
FogControl.fogStartTime = 240
FogControl.fogEndTime = 600
FogControl.lastCheckedDay = -1
FogControl.settingsAppliedForDay = false

function FogControl:debugPrint(str)
    if g_currentMission and g_currentMission.isMasterUser then
        print("[FogControl] " .. str)
    end
end

function FogControl:updateFogSettings()
    if not g_currentMission or not g_currentMission.environment or not g_currentMission.environment.weather then
        return
    end

    local fogUpdater = g_currentMission.environment.weather.fogUpdater
    local fogSettings = FogSettings.new()

    if not FogControl.fogEnabled then
        -- Completely disable fog
        fogSettings.groundFogGroundLevelDensity = 0
        fogSettings.heightFogGroundLevelDensity = 0
        fogSettings.groundFogCoverageEdge0 = 0
        fogSettings.groundFogCoverageEdge1 = 0
        fogSettings.groundFogStartDayTimeMinutes = 0
        fogSettings.groundFogEndDayTimeMinutes = 0
        fogUpdater.visibilityAlpha = 0
        FogControl:debugPrint("Fog disabled: all parameters set to 0")
    else
        -- Apply user settings
        fogSettings.groundFogGroundLevelDensity = FogControl.groundFogDensity
        fogSettings.heightFogGroundLevelDensity = FogControl.heightFogDensity
        fogSettings.groundFogCoverageEdge0 = FogControl.fogCoverage
        fogSettings.groundFogCoverageEdge1 = math.max(FogControl.fogCoverage, FogControl.fogCoverage + 0.01)
        fogSettings.groundFogStartDayTimeMinutes = FogControl.fogStartTime
        fogSettings.groundFogEndDayTimeMinutes = FogControl.fogEndTime
        fogUpdater.visibilityAlpha = 1
        FogControl:debugPrint(string.format("Fog enabled with user settings: groundFogDensity=%.4f, heightFogDensity=%.4f, fogCoverage=%.4f", 
            fogSettings.groundFogGroundLevelDensity, fogSettings.heightFogGroundLevelDensity, fogSettings.groundFogCoverageEdge0))
    end

    fogUpdater:setTargetFog(fogSettings, 30000) -- 30-second transition for smooth fog changes
    fogUpdater.isDirty = true
end

function FogControl:update(dt)
    if not g_currentMission or not g_currentMission:getIsServer() then
        return
    end

    if not FogControl.initialised then
        FogControl:updateFogSettings()
        FogControl.initialised = true
    end

    -- Check if we need to apply fog settings before fogStartTime
    local environment = g_currentMission.environment
    if environment and environment.weather then
        local currentDay = environment.weather.owner.currentMonotonicDay
        local currentDayTime = environment.dayTime / 60000 -- Convert to minutes

        -- Reset settingsAppliedForDay when a new day starts
        if currentDay ~= FogControl.lastCheckedDay then
            FogControl.settingsAppliedForDay = false
            FogControl.lastCheckedDay = currentDay
            FogControl:debugPrint("New day detected, resetting fog settings application flag")
        end

        -- Apply settings 1 minute before fogStartTime if not already applied today
        if not FogControl.settingsAppliedForDay then
            local fogStartTime = FogControl.fogStartTime
            local timeUntilStart = fogStartTime - currentDayTime
            if timeUntilStart > 0 and timeUntilStart <= 1 then -- Within 1 minute before fogStartTime
                FogControl:debugPrint("Applying fog settings before fogStartTime: " .. fogStartTime .. " minutes")
                FogControl:updateFogSettings()
                FogControl.settingsAppliedForDay = true
            end
        end
    end
end

function FogControl:loadMap(name)
    FogControl:readSettings()
    addConsoleCommand("gsFogControl", "Control fog settings", "consoleCommandFogControl", FogControl)
    -- Apply settings immediately after map load to override environment.xml
    FogControl:debugPrint("Map loaded, applying initial fog settings")
    FogControl:updateFogSettings()
end

function FogControl:onDayChanged()
    -- Reapply fog settings when a new day starts
    FogControl:debugPrint("New day detected, reapplying fog settings")
    FogControl:updateFogSettings()
end

-- Console command
function FogControl:consoleCommandFogControl(param1, param2)
    if param1 == "enable" then
        local value = param2:lower() == "true"
        FogControl.fogEnabled = value
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Fog enabled: %s", tostring(value))
    elseif param1 == "groundDensity" then
        local value = tonumber(param2) or FogControl.groundFogDensity
        FogControl.groundFogDensity = math.clamp(value, 0.0, 1.0)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Ground fog density: %.4f", FogControl.groundFogDensity)
    elseif param1 == "heightDensity" then
        local value = tonumber(param2) or FogControl.heightFogDensity
        FogControl.heightFogDensity = math.clamp(value, 0.0, 1.0)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Height fog density: %.4f", FogControl.heightFogDensity)
    elseif param1 == "coverage" then
        local value = tonumber(param2) or FogControl.fogCoverage
        FogControl.fogCoverage = math.clamp(value, 0.0, 1.0)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Fog coverage: %.4f", FogControl.fogCoverage)
    elseif param1 == "startTime" then
        local value = tonumber(param2) or FogControl.fogStartTime
        FogControl.fogStartTime = math.clamp(value, 0, 1440)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Fog start time: %d minutes", FogControl.fogStartTime)
    elseif param1 == "endTime" then
        local value = tonumber(param2) or FogControl.fogEndTime
        FogControl.fogEndTime = math.clamp(value, 0, 1440)
        FogControl:writeSettings()
        FogControl:updateFogSettings()
        return string.format("Fog end time: %d minutes", FogControl.fogEndTime)
    end
    return "Usage: gsFogControl [enable|groundDensity|heightDensity|coverage|startTime|endTime] <value>"
end

-- Settings management
function FogControl:setValue(id, value)
    FogControl[id] = value
end

function FogControl:getValue(id)
    return FogControl[id]
end

function FogControl:getStateIndex(id, value)
    local value = value or FogControl:getValue(id)
    local setting = FogControl.SETTINGS[id]
    if not setting then
        return 1
    end
    local values = setting.values
    if type(value) == 'number' then
        local index = setting.default
        local initialDiff = math.huge
        for i, v in pairs(values) do
            local currentDiff = math.abs(v - value)
            if currentDiff < initialDiff then
                initialDiff = currentDiff
                index = i
            end
        end
        return index
    else
        for i, v in pairs(values) do
            if v == value then
                return i
            end
        end
    end
    return setting.default
end

function FogControl:writeSettings()
    local key = "fogControlSettings"
    local userSettingsFile = Utils.getFilename("modSettings/FogControl.xml", getUserProfileAppPath())
    local xmlFile = createXMLFile("settings", userSettingsFile, key)

    if xmlFile ~= 0 then
        for _, id in pairs(FogControl.menuItems) do
            local value = FogControl:getValue(id)
            local xmlValueKey = "fogControlSettings." .. id .. "#value"
            if type(value) == 'number' then
                setXMLFloat(xmlFile, xmlValueKey, value)
            elseif type(value) == 'boolean' then
                setXMLBool(xmlFile, xmlValueKey, value)
            end
        end
        saveXMLFile(xmlFile)
        delete(xmlFile)
    end
end

function FogControl:readSettings()
    local userSettingsFile = Utils.getFilename("modSettings/FogControl.xml", getUserProfileAppPath())
    if not fileExists(userSettingsFile) then
        FogControl:writeSettings()
        return
    end

    local xmlFile = loadXMLFile("fogControlSettings", userSettingsFile)
    if xmlFile ~= 0 then
        for _, id in pairs(FogControl.menuItems) do
            local setting = FogControl.SETTINGS[id]
            if setting then
                local xmlValueKey = "fogControlSettings." .. id .. "#value"
                if hasXMLProperty(xmlFile, xmlValueKey) then
                    local value = FogControl:getValue(id)
                    if type(value) == 'number' then
                        value = getXMLFloat(xmlFile, xmlValueKey) or value
                    elseif type(value) == 'boolean' then
                        value = getXMLBool(xmlFile, xmlValueKey) or value
                    end
                    FogControl:setValue(id, value)
                end
            end
        end
        delete(xmlFile)
    end
end

-- Menu integration
FogControl.SETTINGS.fogEnabled = {
    default = 1,
    values = {false, true},
    strings = {g_i18n:getText("ui_off"), g_i18n:getText("ui_on")}
}

-- Generate percentage values for all parameters (0% to 100% with 0.1% step)
local percentValues = {}
local percentStrings = {}
for i = 0, 1000 do
    percentValues[i + 1] = i / 1000 -- 0.000 to 1.000 with step 0.001
    percentStrings[i + 1] = string.format("%.1f%%", i / 10)
end

FogControl.SETTINGS.groundFogDensity = {
    default = 5, -- 0.5% (0.005)
    values = percentValues,
    strings = percentStrings
}

FogControl.SETTINGS.heightFogDensity = {
    default = 600, -- 60% (0.6)
    values = percentValues,
    strings = percentStrings
}

FogControl.SETTINGS.fogCoverage = {
    default = 15, -- 1.5% (0.015)
    values = percentValues,
    strings = percentStrings
}

FogControl.SETTINGS.fogStartTime = {
    default = 2,
    values = {0, 240, 300, 360, 420},
    strings = {"00:00", "04:00", "05:00", "06:00", "07:00"}
}

FogControl.SETTINGS.fogEndTime = {
    default = 2,
    values = {480, 540, 600, 660, 720},
    strings = {"08:00", "09:00", "10:00", "11:00", "12:00"}
}

function FogControl:addMenuOption(id)
    local inGameMenu = g_gui.screenControllers[InGameMenu]
    local settingsPage = inGameMenu.pageSettings
    local settingsLayout = settingsPage.generalSettingsLayout

    local original = #FogControl.SETTINGS[id].values == 2 and settingsPage.checkWoodHarvesterAutoCutBox or settingsPage.multiVolumeVoiceBox
    local options = FogControl.SETTINGS[id].strings

    local menuOptionBox = original:clone(settingsLayout)
    menuOptionBox.id = id .. "box"

    local menuOption = menuOptionBox.elements[1]
    menuOption.id = id
    menuOption.target = FogControl
    menuOption:setCallback("onClickCallback", "onMenuOptionChanged")
    menuOption:setDisabled(false)

    local toolTip = menuOption.elements[1]
    toolTip:setText(g_i18n:getText("tooltip_fogcontrol_" .. id))

    local setting = menuOptionBox.elements[2]
    setting:setText(g_i18n:getText("setting_fogcontrol_" .. id))

    menuOption:setTexts({unpack(options)})
    menuOption:setState(FogControl:getStateIndex(id))

    FogControl.CONTROLS[id] = menuOption

    local function updateFocusIds(element)
        if not element then return end
        element.focusId = FocusManager:serveAutoFocusId()
        for _, child in pairs(element.elements) do
            updateFocusIds(child)
        end
    end
    updateFocusIds(menuOptionBox)
    table.insert(settingsPage.controlsList, menuOptionBox)
end

function FogControl:onMenuOptionChanged(state, menuOption)
    local id = menuOption.id
    local value = FogControl.SETTINGS[id].values[state]
    FogControl:setValue(id, value)
    FogControl:writeSettings()
    FogControl:updateFogSettings()
end

-- Add menu section and options
local inGameMenu = g_gui.screenControllers[InGameMenu]
local settingsPage = inGameMenu.pageSettings
local settingsLayout = settingsPage.generalSettingsLayout

local sectionTitle
for _, elem in ipairs(settingsLayout.elements) do
    if elem.name == "sectionHeader" then
        sectionTitle = elem:clone(settingsLayout)
        break
    end
end
if sectionTitle then
    sectionTitle:setText(g_i18n:getText("menu_FogControl_TITLE"))
else
    local title = TextElement.new()
    title:applyProfile("fs25_settingsSectionHeader", true)
    title:setText(g_i18n:getText("menu_FogControl_TITLE"))
    title.name = "sectionHeader"
    settingsLayout:addElement(title)
    sectionTitle = title
end

sectionTitle.focusId = FocusManager:serveAutoFocusId()
table.insert(settingsPage.controlsList, sectionTitle)
FogControl.CONTROLS[sectionTitle.name] = sectionTitle

for _, id in pairs(FogControl.menuItems) do
    FogControl:addMenuOption(id)
end
settingsLayout:invalidateLayout()

-- Handle focus for custom controls
FocusManager.setGui = Utils.appendedFunction(FocusManager.setGui, function(_, gui)
    if gui == "ingameMenuSettings" then
        for _, control in pairs(FogControl.CONTROLS) do
            if not control.focusId or not FocusManager.currentFocusData.idToElementMapping[control.focusId] then
                if not FocusManager:loadElementFromCustomValues(control, nil, nil, false, false) then
                    Logging.warning("Could not register control %s with the focus manager", control.id or control.name or control.focusId)
                end
            end
        end
        local settingsPage = g_gui.screenControllers[InGameMenu].pageSettings
        settingsPage.generalSettingsLayout:invalidateLayout()
    end
end)

InGameMenuSettingsFrame.onFrameOpen = Utils.appendedFunction(InGameMenuSettingsFrame.onFrameOpen, function()
    local isAdmin = g_currentMission:getIsServer() or g_currentMission.isMasterUser
    for _, id in pairs(FogControl.menuItems) do
        local menuOption = FogControl.CONTROLS[id]
        menuOption:setState(FogControl:getStateIndex(id))
        menuOption:setDisabled(not isAdmin)
    end
end)

-- Register for day change event
Environment.onDayChanged = Utils.appendedFunction(Environment.onDayChanged, FogControl.onDayChanged)